/*

    module to define and use macros within apache runtime configuration files.
    Copyright (C) 1998 Fabien COELHO

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

 */

/*
  $Id: mod_macro.c,v 1.3 1998/12/16 14:22:18 coelho Exp $

  $Log: mod_macro.c,v $
  Revision 1.3  1998/12/16 14:22:18  coelho
  moved to GNU GPL v2.
  Discussion moved to index.html.

  Revision 1.2  1998/12/16 12:58:42  coelho
  case-insensitive comparisons.

  Revision 1.1  1998/12/15 18:06:41  coelho
  Initial revision

*/

#include "httpd.h"
#include "http_config.h"

/************************************************ COMPILE TIME DEBUG CONTROL */
/*
  #define DEBUG_MODULE_MACRO 1

  CFLAGS="-g -Wall" \
  ./configure --prefix=/users/cri/coelho/tmp/apache \
              --activate-module=src/modules/extra/mod_macro.c

  run -f /users/cri/coelho/tmp/apache/etc/httpd.conf 
*/

#if defined(DEBUG_MODULE_MACRO)
#define debug(stmt) stmt
#else
#define debug(stmt) 
#endif

/********************************************************** MACRO MANAGEMENT */

module macro_module;

/* a macro: name, arguments, contents...
 */
typedef struct {
    char * name;
    array_header /* of char* */ * arguments;
    array_header /* of char* */ * lines;
    int processing; /* whether the macro is being processed. */
    char * location; /* of the macro definition. */
} macro_t;

/* configuration tokens.
 */
#define BEGIN_MACRO "<Macro"
#define END_MACRO "</Macro>"
#define USE_MACRO "Use"

#define empty_string_p(p) (!(p) || *(p)=='\0')

/* macros are kept globally...
   they are not per-server or per-directory entities.

   I would need a hook BEFORE and AFTER configuration processing
   to initialize and close them properly.
   
   I would have such a hook at main/ I guess.

   The "initializer" does not seem to be called before.

   HACK: 
   -----
   I put them in the temp_pool.
   on restarts: I detect them because the temp_pool has changed.
   note that there is always restart implicitely to check for the syntax.
 */
static array_header * macros = NULL;

/* returns the macro structure for name, or NULL if not found.
 */
static macro_t * get_macro_by_name(array_header * macros, char * name)
{
    int i;

    ap_assert(macros);

    for (i=0; i<macros->nelts; i++) {
	macro_t * macro = ((macro_t **) macros->elts)[i];
	if (!strcasecmp(name, macro->name)) {
	    return macro;
	}
    }
    return NULL;
}

/* configuration state initialization.
   the state is simply an array_header which holds the macros.
 */
static void macro_init(pool * p)
{
    static ap_pool * last_time_temp_pool_hack = NULL;

    debug(fprintf(stderr, "macro_init\n"));

    /* is it a restart? */
    if (last_time_temp_pool_hack!=p)
    {
	last_time_temp_pool_hack = p;
	macros = ap_make_array(p, 1, sizeof(macro_t *));
	debug(fprintf(stderr, "macro_init done for %p\n", p));
    }
}

/*************************************************************** PARSE UTILS */

#define trim(line) while (*(line)==' ' || *(line)=='\t') (line)++

/* return configuration-parsed arguments from line as an array. 
 */
static array_header * get_arguments(ap_pool * p, const char * line)
{
    array_header * args = ap_make_array(p, 1, sizeof(char*));
    char * arg, ** new;

    trim(line);
    while (*line) {
	arg = ap_getword_conf(p, &line);
	new = ap_push_array(args);
	*new = arg;
	trim(line);
    }

    return args;
}

/* get read lines as an array till end_token. 
   counts nesting. this function could be exported.
 */
static char * get_lines_till_end_token(
    ap_pool * p, 
    configfile_t * config_file, 
    const char * end_token,
    const char * begin_token, /* may be NULL. */
    array_header ** plines)
{
    array_header * lines = ap_make_array(p, 1, sizeof(char*));
    char ** new, * first, * ptr;
    char line[MAX_STRING_LEN];
    int nesting = 1;

    while (!ap_cfg_getline(line, MAX_STRING_LEN, config_file)) {
	ptr = line;
	first = ap_getword_conf_nc(p, &ptr);
	if (first) {
	    if (!strcasecmp(first, end_token)) { /* okay! */
		nesting--;
		if (!nesting) {
		    *plines = lines;
		    return NULL;
		}
	    } 
	    else if (begin_token && !strcasecmp(first, begin_token)) {
		nesting++;
	    }
	}
	/* free first. */
	new  = ap_push_array(lines);
	*new = ap_psprintf(p, "%s\n", line); /* put '\n' back */
    }

    return ap_pstrcat(p, "expected token not found: ", end_token, NULL);
}

/******************************************************** SUBSTITUTION UTILS */

/* replace name by replacement at the beginning of buf of bufsize.
   returns an error message or NULL.
 */
static char * substitute(
    char * buf, int bufsize,
    char * name, char * replacement)
{
    char tmp[MAX_STRING_LEN];
    int lbuf = strlen(buf), 
	lname = strlen(name), 
	lrepl = strlen(replacement);

    if (lbuf + lrepl - lname >= bufsize) {
	return "cannot substitute, buffer size too small";
    } else if (lbuf + lrepl - lname >= MAX_STRING_LEN) {
	return "cannot substitute, tmp size too small";
    }

    /* simple. */
    strcpy(tmp, replacement);
    strcpy(tmp+lrepl, buf+lname);
    strcpy(buf, tmp);

    return NULL;
}

/* find first occurence of args in buf.
   returns the pointer and the whichone found, or NULL.
 */
static char * next_substitution(char * buf, array_header * args, int *whichone)
{
    int i;
    char * first = NULL, * found;

    for (i=0; i<args->nelts; i++) {
	found = strstr(buf, ((char**)args->elts)[i]);
	if (found && (!first || found<first)) {
	    first = found;
	    *whichone = i;
	}
    }

    return first;
}

/* substitute macro arguments by replacements in buf of bufsize.
   returns an error message or NULL;
 */
static char * substitute_macro_args(
    char * buf, int bufsize, macro_t * macro, array_header * replacements)
{
    int whichone;
    char * ptr = buf, * errmsg;

    while ((ptr = next_substitution(ptr, macro->arguments, &whichone))) {
	errmsg = substitute(ptr, buf-ptr+bufsize, 
			    ((char**)macro->arguments->elts)[whichone],
			    ((char**)replacements->elts)[whichone]);
	ptr += strlen(((char**)replacements->elts)[whichone]);
	if (errmsg) return errmsg;
    }

    return NULL;
}

/* perform substitutions in a macro contents and return the result as
   an array. may also return an error message.
 */
static const char * process_content(
    ap_pool * p,				    
    macro_t * macro, 
    array_header * replacements,
    array_header ** result)
{
    array_header * lines = macro->lines;
    char line[MAX_STRING_LEN], ** new, * errmsg;
    int i;
    
    *result = ap_make_array(p, 1, sizeof(char*));
    
    for (i=0; i<lines->nelts; i++) {
	strncpy(line, ((char**)lines->elts)[i], MAX_STRING_LEN-1);
	errmsg = substitute_macro_args(line, MAX_STRING_LEN, 
				       macro, replacements);
	if (errmsg) {
	    return ap_pstrcat(p, "while processing macro ", macro->name, 
			      " ", errmsg, NULL);
	}
	
	new = ap_push_array(*result);
	*new = ap_pstrdup(p, line);
    }

    return NULL;
}

/********************************************************* MACRO CONFIG FILE */

/* the expanded content of the macro is to be parsed as a configfile_t. 
   the following struct stores the content.


   IMPORTANT NOTE:
   ---------------

   in http_config.c there is such a stuff made static,
   which does not implement getch().
   maybe this should be moved to util.c ???
 */
typedef struct {
    int index;               /* current element. */
    int char_index;          /* current char in element. */
    int length;              /* cached length of the current line. */
    array_header * contents; /* array of char * */
} array_lines_t;

/* returns next char or -1.
 */
static int array_getch(void * param)
{
    array_lines_t * ml = (array_lines_t*) param;

    while (ml->char_index>=ml->length) { /* next element */
	if (ml->index>=ml->contents->nelts) {
	    return -1;
	}
	ml->index++;
	ml->char_index = 0;
	ml->length = ml->index>=ml->contents->nelts? 0: 
	    strlen(((char**)ml->contents->elts)[ml->index]);
    }

    return (((char**)ml->contents->elts)[ml->index])[ml->char_index++];
}

/* returns a buf a la fgets.
   no more than a line at a time, otherwise the parsing is too much ahead...
   NULL at EOF.
 */
static void * array_getstr(void * buf, size_t bufsize, void * param)
{
    char * buffer = (char *) buf;
    int i=0, next=0;

    while (i<bufsize-1 && next!='\n' && ((next=array_getch(param))!=-1)) {
	buffer[i++] = (char) next;
    }
    
    if (next==-1 && i==0) return NULL; /* EOF */

    buffer[i] = '\0';
    
    debug(fprintf(stderr, "array_getstr: returning: ---%s---\n", buf));
    return buf;
}

/* close the array stream.
 */
static int array_close(void * param)
{
    array_lines_t * ml = (array_lines_t*) param;
    ml->index = ml->contents->nelts;
    ml->char_index = ml->length;
    return 0;
}

/* this one could be exported.
 */
static configfile_t * make_array_config(
    ap_pool * p,
    array_header * contents,
    const char * where)
{
    array_lines_t * ls = (array_lines_t *) ap_palloc(p, sizeof(array_lines_t));
    
    ls->index      = 0;
    ls->char_index = 0;
    ls->contents   = contents;
    ls->length     = ls->contents->nelts<1? 0: 
	strlen(((char**)ls->contents->elts)[0]);

    return ap_pcfg_open_custom(p, where, (void*) ls,
			       array_getch, array_getstr, array_close);
}


/********************************************************** KEYWORD HANDLING */

/* handles <Macro macroname macroarg1 macroarg2 ...>
 */
static const char * macro_section(cmd_parms *cmd, void *dummy, const char *arg)
{
    const char * errmsg;
    char * endp = strrchr(arg, '>'), ** new;
    macro_t 
	* macro = (macro_t*) ap_palloc(cmd->temp_pool, sizeof(macro_t)),
	* old;

    debug(fprintf(stderr, "macro_section -%s-\n", arg));

    macro_init(cmd->temp_pool); /* lazy... */

    if (endp) *endp = '\0';

    /* get name. */
    macro->name = ap_getword_conf(cmd->temp_pool, &arg);
    if (empty_string_p(macro->name)) {
	return "macro definition: name not specified";
    }

    old = get_macro_by_name(macros, macro->name);
    if (old) {
	return ap_psprintf(cmd->temp_pool, 
			   "macro %s multiply defined. already seen: %s.", 
			   old->name, old->location);
    }

    /* get arguments. */
    macro->arguments = get_arguments(cmd->temp_pool, arg);
    macro->location = ap_psprintf(cmd->temp_pool, 
				  "defined at %s:%d",
				  cmd->config_file->name,
				  cmd->config_file->line_number);

    /* get contents */
    errmsg = get_lines_till_end_token(cmd->temp_pool, cmd->config_file,
				      END_MACRO, BEGIN_MACRO, &macro->lines);

    if (errmsg) 
	return ap_psprintf(cmd->temp_pool, 
			   "macro %s %s\n\tcontents error\n%s",
			   macro->name, macro->location, errmsg);
    
    /* add the new macro. */
    macro->processing = 0;
    new  = ap_push_array(macros);
    *new = (char *) macro;

    return NULL;
}

/* handles: Use name arg1 ...
 */
static const char * use_macro(cmd_parms * cmd, void * dummy, const char * arg)
{
    cmd_parms params;
    char * name, * where;
    const char * errmsg;
    array_header * contents, * replacements;
    macro_t * macro;

    debug(fprintf(stderr, "use_macro -%s-\n", arg));

    macro_init(cmd->temp_pool); /* lazy... */

    name = ap_getword_conf(cmd->temp_pool, &arg);
    macro = get_macro_by_name(macros, name);
    replacements = get_arguments(cmd->temp_pool, arg);

    if (!macro) {
	return ap_psprintf(cmd->temp_pool, "no %s macro defined", name);
    }

    if (macro->arguments->nelts!=replacements->nelts) {
	return ap_psprintf(cmd->temp_pool, 
	    "use of macro %s with %d argument%s instead of %d",
			   name, replacements->nelts,
			   replacements->nelts>1? "s": "", /* (grammar;-) */
			   macro->arguments->nelts);
    }

    if (macro->processing) {
	return ap_psprintf(cmd->temp_pool, "macro %s is being processed. "
			   "recursion not allowed.", name);
    }

    macro->processing = 1; /* detect recursion. */

    /* the following is similar to the static process_command_config 
       in http_config.h. maybe some common parts should be extracted.
     */
    where = ap_psprintf(cmd->temp_pool, "macro %s used at %s:%d", 
			macro->name, 
			cmd->config_file->name, cmd->config_file->line_number);
    
    errmsg = process_content(cmd->temp_pool, macro, replacements, &contents);

    if (errmsg) 
	return ap_psprintf(cmd->temp_pool, 
			   "%s substitution error\n\t%s\n%s", 
			   where, macro->location, errmsg);

    params.pool        = cmd->pool;
    params.temp_pool   = cmd->temp_pool;
    params.server      = cmd->server;
    params.override    = cmd->override;
    params.limited     = cmd->limited;
    params.info        = cmd->info;
    params.end_token   = cmd->end_token; /* ??? */
    params.cmd         = cmd->cmd;
    params.path        = cmd->path;
    params.config_file = make_array_config(cmd->temp_pool, contents, where);
    
    errmsg = ap_srm_command_loop(&params, params.server->lookup_defaults);

    if (errmsg) 
	return ap_psprintf(cmd->temp_pool, 
			   "%s now line %d\n\t%s\n"
			   "\texecution error:\n%s", 
			   where, 
			   params.config_file->line_number,
			   macro->location, errmsg);

    macro->processing = 0;
    return NULL;
}

/* handles a lonely </Macro>
 */
static const char * unexpected_end_macro(
    cmd_parms * cmd, void * dummy, const char * arg)
{
    return ap_psprintf(cmd->temp_pool, "unexpected " END_MACRO);
}

/* static char * hello(void) { fprintf(stderr, "hello\n"); return NULL; }
 */

/************************************************************* EXPORT MODULE */

static const command_rec macro_cmds[] =
{
    { BEGIN_MACRO, macro_section, NULL, RSRC_CONF, RAW_ARGS,
      "Contains a macro within the configuration file." },
    { END_MACRO, unexpected_end_macro, NULL, RSRC_CONF, NO_ARGS,
      "End of a macro defition." },
    { USE_MACRO, use_macro, NULL, RSRC_CONF, RAW_ARGS,
      "Use of a macro." },
    { NULL }
};

module MODULE_VAR_EXPORT macro_module =
{
    STANDARD_MODULE_STUFF,
    NULL,	                /* initializer */
    NULL,                       /* create per-dir config */
    NULL,                       /* merge per-dir config */
    NULL,                       /* server config */
    NULL,                       /* merge server config */
    macro_cmds,                 /* command table */
    NULL,                       /* handlers */
    NULL,                       /* filename translation */
    NULL,                       /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    NULL,                       /* child_init */
    NULL,                       /* child_exit */
    NULL                        /* post read-request */
};
